/*
 ************************************************************************************************
 *                                                                                              
 *  OSA cooperative RTOS for microcontrollers PIC (Microchip) and 8-bit AVR (Atmel)
 *                                                                                              
 *  OSA is distributed under BSD license (see license.txt)                                      
 *                                                                                              
 *  URL:        http://wiki.pic24.ru/doku.php/en/osa/ref/intro                                  
 *              http://picosa.narod.ru                                                          
 *                                                                                              
 *----------------------------------------------------------------------------------------------
 *                                                                                              
 *  File:       osa_system.h                                                                    
 *                                                                                              
 *  Programmer: Timofeev Victor                                                                 
 *              osa@pic24.ru, testerplus@mail.ru                                                
 *                                                                                              
 *  Definition: System services                                                                 
 *                                                                                              
 *  History:    21.01.2009                                                                      
 *              03.11.2009 -    Bug fixed: _OS_CHECK_EVENT and _OS_CHECK_EVENT_TO are called    
 *                              now with (event)!=0 (instead of event), since parameter has     
 *                              type of char (some condition could be for example = 0xFF00)
 *              06.11.2009 -    Macros OS_Wait and OS_Wait_TO now defined with parameters, since
 *                              mikroC PRO exited with an error
 *                                                                                              
 ************************************************************************************************
 */














/************************************************************************************************
 *                                                                                              *
 *                                      S H C E D U L E R                                       *
 *                                                                                              *
 *                      !!!(Sheduler for MCC30 is in "port\osa_mcc30.h")!!!                     *
 *                                                                                              *
 ************************************************************************************************/


//------------------------------------------------------------------------------
#ifndef OS_Sched
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
#ifdef OS_ENABLE_CRITICAL_SECTION

    #define SKIP_IF_IS_IN_CRITICAL_SECTION()    if (OS_IsInCriticalSection()) goto SCHED_CHECK_READY;
    #define LABEL_SCHED_CHECK_READY             SCHED_CHECK_READY:
    #define NOT_IN_CRITICAL_SECTION()           !OS_IsInCriticalSection()

#else

    #define SKIP_IF_IS_IN_CRITICAL_SECTION()
    #define LABEL_SCHED_CHECK_READY
    #define NOT_IN_CRITICAL_SECTION()           1

#endif
//------------------------------------------------------------------------------



#define LABEL_SCHED_CONTINUE            _LABEL_SCHED_CONTINUE:;

#define IF_NOT_TASK_ENABLE_CONTINUE()   if (!_OS_bTaskEnable || _OS_bTaskPaused) \
                                            goto _LABEL_SCHED_CONTINUE;

#define IF_NOT_TASK_READY()             if (!_OS_bTaskReady)

#define IF_TASK_READY()                 if (_OS_bTaskReady)

#define LABEL_SCHED_SKIP_RUN


#ifdef OS_ENABLE_TTIMERS

    #define IF_NOT_OS_CHECK_DELAYS_GOTO_SCHED_CONTINUE                                      \
                                        if (_OS_bTaskDelay && !_OS_bTaskCanContinue)        \
                                            goto _LABEL_SCHED_CONTINUE;


#else

    #define IF_NOT_OS_CHECK_DELAYS_GOTO_SCHED_CONTINUE


#endif



//------------------------------------------------------------------------------
#if defined(_OS_CUR_FLAGS_IN_OS_state)
//------------------------------------------------------------------------------

    #define _OS_SET_STATE()     OS_state = _OS_CurTask->State
    #define _OS_RESTORE_INDF()

//------------------------------------------------------------------------------
#else
//------------------------------------------------------------------------------

    #define _OS_SET_STATE()     _OS_SET_FSR_CUR_TASK()
    #define _OS_RESTORE_INDF()  _OS_SET_FSR_CUR_TASK()

//------------------------------------------------------------------------------
#endif
//------------------------------------------------------------------------------




//------------------------------------------------------------------------------
#if !defined(OS_DISABLE_PRIORITY)
//------------------------------------------------------------------------------

    /********************************************************************************************
     *                                                                                          *
     *     P R I O R I T Y   M O D E   S C H E D U L E R                                        *
     *                                                                                          *
     ********************************************************************************************/

    #define OS_IsIdle()     (!OS_IsInCriticalSection() && !OS_Flags.bBestTaskFound)

    #define OS_Sched()                                                                          \
    {                                                                                           \
                                                                                                \
        /* First we suppose that there is no ready task*/                                       \
        OS_Flags.bBestTaskFound = 0;                                                            \
                                                                                                \
        /* MODE: Search of best task                  */                                        \
        OS_Flags.bCheckingTasks = 1;                                                            \
                                                                                                \
        /* Worst of lowest priority                   */                                        \
        _OS_best_priority = OS_WORST_PRIORITY + 1;                                              \
                                                                                                \
        SKIP_IF_IS_IN_CRITICAL_SECTION();                                                       \
                                                                                                \
        /* List all tasks starting from last executed   */                                      \
        _OS_CurTask = _OS_LastTask;                                                             \
                                                                                                \
        do                                                                                      \
        {                                                                                       \
            if (_OS_CurTask >= (OST_TASK_POINTER)OS_TaskVars + (OS_TASKS-1))                    \
                _OS_CurTask = (OST_TASK_POINTER)OS_TaskVars;                                    \
            else                                                                                \
                _OS_CurTask++;                                                                  \
                                                                                                \
    LABEL_SCHED_CHECK_READY                                                                     \
                                                                                                \
            _OS_SET_STATE();                                                                    \
                                                                                                \
            IF_NOT_TASK_ENABLE_CONTINUE();                           /* Skip empty descriptor */\
                                                                                                \
            if (_OS_best_priority <= (_OS_temp = _OS_cTaskPriority)) /* see p.1 below         */\
                goto _LABEL_SCHED_CONTINUE;                                                     \
                                                                                                \
            IF_NOT_TASK_READY()                                                                 \
            {                                                                                   \
                /* Task is not ready yet. Check for it has became ready    */                   \
    SCHED_RUN:;                                                                                 \
                _OS_JumpToTask();                                                               \
                _OS_SaveRetAddr();                                                              \
                if (!OS_Flags.bCheckingTasks) goto SCHED_END;                                   \
                _OS_RESTORE_INDF();                                                             \
            }                                                                                   \
                                                                                                \
    LABEL_SCHED_SKIP_RUN                                                                        \
                                                                                                \
            /* Skip if task is delayed              */                                          \
            IF_NOT_OS_CHECK_DELAYS_GOTO_SCHED_CONTINUE;                                         \
                                                                                                \
            /* Compare priority for all ready tasks      */                                     \
            IF_TASK_READY()                                                                     \
            {                                                                                   \
                OS_Flags.bBestTaskFound = 1;                                                    \
                _OS_BestTask = _OS_CurTask;                                                     \
                _OS_best_priority = _OS_cTaskPriority;                                          \
                                                                                                \
                /* If this task has highest priority then stop searching */                     \
                if (_OS_best_priority == OS_BEST_PRIORITY) break;                               \
            }                                                                                   \
                                                                                                \
    LABEL_SCHED_CONTINUE                                                                        \
                                                                                                \
        } while (NOT_IN_CRITICAL_SECTION() && _OS_CurTask != _OS_LastTask);                     \
                                                                                                \
        /* If there is at least one ready task, then execute it */                              \
                                                                                                \
        if (OS_Flags.bBestTaskFound)                                                            \
        {                                                                                       \
            OS_Flags.bCheckingTasks = 0;                /* MODE: executing best task    */      \
            _OS_LastTask = _OS_CurTask = _OS_BestTask;                                          \
            goto SCHED_RUN;                                                                     \
        }                                                                                       \
                                                                                                \
    SCHED_END:;                                                                                 \
                                                                                                \
    }                                                                                           \

    //-------------------------------------------------------------------------------------------
    // Explaining of some moments:
    //
    // 1.  OS_Temp used here to reduce used RAM by 1 byte. Compiler while comparing
    //     _OS_best_priority with (indf&7) generates one temporary variable. We set it manualy,
    //     so compiler will not use local variable.
    //
    //-------------------------------------------------------------------------------------------



//------------------------------------------------------------------------------
#endif      // !OS_DISABLE_PRIORITY
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
#if defined(OS_DISABLE_PRIORITY)
//------------------------------------------------------------------------------

    /********************************************************************************************
     *                                                                                          *
     *     N O N   P R I O R I T Y   M O D E   S C H E D U L E R                                *
     *                                                                                          *
     ********************************************************************************************/

    #define OS_IsIdle()     0

    //-------------------------------------------------------------------------------------------
    //
    // Non priority mode can be used when there is no need to give more preference to
    // some tasks. (For example there are no time rcitical code). This mode allows to reduce ROM
    // and RAM usage and also increases speed of // scheduler work in OS_TASKS times
    //
    //-------------------------------------------------------------------------------------------


    #ifdef OS_ENABLE_CRITICAL_SECTION
        #define IF_NOT_IS_IN_CRITICAL_SECTION()        if (!OS_IsInCriticalSection())
    #else
        #define IF_NOT_IS_IN_CRITICAL_SECTION()
    #endif

    #define OS_Sched()                                                                          \
    {                                                                                           \
        _OS_temp = 0;                                                                           \
                                                                                                \
        IF_NOT_IS_IN_CRITICAL_SECTION()                                                         \
        {                                                                                       \
            if (_OS_CurTask >= (OST_TASK_POINTER)OS_TaskVars + (OS_TASKS-1))                    \
                _OS_CurTask = (OST_TASK_POINTER)OS_TaskVars;                                    \
            else                                                                                \
                _OS_CurTask++;                                                                  \
        }                                                                                       \
                                                                                                \
        _OS_SET_STATE();                                                                        \
                                                                                                \
        /* Skip empty descriptor              */                                                \
        if (!_OS_bTaskEnable || _OS_bTaskPaused) goto _LABEL_SCHED_CONTINUE;                    \
                                                                                                \
        IF_NOT_OS_CHECK_DELAYS_GOTO_SCHED_CONTINUE;                                             \
                                                                                                \
        _OS_JumpToTask();                                                                       \
        _OS_SaveRetAddr();                                                                      \
                                                                                                \
    _LABEL_SCHED_CONTINUE:;                                                                     \
                                                                                                \
    }

//------------------------------------------------------------------------------
#endif     //   OS_DISABLE_PRIORITY
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
#endif  // ifndef OS_Shed
//------------------------------------------------------------------------------

/************************************************************************************************/









/************************************************************************************************
 *                                                                                              *
 *     C O M M O N   S Y S T E M   S E R V I C E S                                              *
 *                                                                                              *
 ************************************************************************************************/


#define OS_Run()    for(;;) OS_Sched()


/************************************************************************************
 *
 * Give control to scheduler:
 *
 ************************************************************************************/


/************************************************************************/
/* 1. Return with saving context                                        */
/************************************************************************/
#define OS_Yield()                          _OS_RETURN_SAVE()

/************************************************************************/
/* 2. Return without saving context                                     */
/*    (internal service)                                                */
/************************************************************************/
#define _OS_Return()                        _OS_RETURN_NO_SAVE()

/************************************************************************/
/* 3. Clear bReady and save context without returning                   */
/************************************************************************/
#define _OS_ClearReady()                    _OS_CLEAR_READY()

/************************************************************************/
/* 4. Clear bReady, set bCanContinue, save context without returning    */
/************************************************************************/
#define _OS_ClearReadySetCanContinue()      _OS_CLEAR_READY_SET_CANCONTINUE()












/************************************************************************/
/* Check for timeout expired in last service                            */
/************************************************************************/

#define OS_IsTimeout()              OS_Flags.bTimeout

/************************************************************************/
/* Check for errors:                                                    */
/************************************************************************/

//------------------------------------------------------------------------------
// Check for task creation error

#define OS_IsError()                OS_Flags.bError

//------------------------------------------------------------------------------
// Check for event error

#define OS_IsEventError()           OS_Flags.bEventError

/************************************************************************/
/* Check for one of task is in critical section                         */
/************************************************************************/

#define OS_IsInCriticalSection()    OS_Flags.bInCriticalSection













/************************************************************************************************
 *                                                                                              *
 *                                    I N T E R R U P T S                                       *
 *                                                                                              *
 ************************************************************************************************/


//------------------------------------------------------------------------------
// Enable/disable/restore interrupts are in h-files for each processor
//------------------------------------------------------------------------------



//------------------------------------------------------------------------------
#if defined(__OSAPICC16__) || defined(__OSAPICC18__)    /*100311.10*/
//------------------------------------------------------------------------------

    /************************************************************************/
    /*                                                                      */
    /* This service must be called in the begin of ISR                      */
    /*                                                                      */
    /************************************************************************/

    #define OS_EnterInt()                                   \
        static _FSR_TYPE   _OS_FSR_Temp;                    \
        _OS_FSR_Temp = _fsr;                                \

    /************************************************************************/
    /*                                                                      */
    /* This service must be called at the end of ISR                        */
    /*                                                                      */
    /************************************************************************/

    #define OS_LeaveInt()   _fsr = _OS_FSR_Temp

//------------------------------------------------------------------------------
#else
//------------------------------------------------------------------------------

    #define OS_EnterInt()
    #define OS_LeaveInt()

#endif







/************************************************************************************************
 *                                                                                              *
 *                                     C O N D I T I O N S                                      *
 *                                                                                              *
 ************************************************************************************************/



/***********************************************************************************************
 *
 * Two internal macros for reducing source code of operation system.
 *
 * !!! FOR AUTHOR: You should be careful with using these macros because of "event"
 * is any expression including modifying FSR.
 *
 ***********************************************************************************************/

/*
 ********************************************************************************
 *                                                                              *
 *   Put task in wait mode till "event" will not became true                    *
 *                                                                              *
 ********************************************************************************
 */

#define _OS_WAIT_EVENT(event)                       \
{                                                   \
    _OS_ClearReady();                               \
    _OS_CHECK_EVENT((event) != 0);      /*91103*/   \
}

/*
 ********************************************************************************
 *                                                                              *
 *   Put task in wait mode till "event" will not became true or timeout will not*
 *   expired                                                                    *
 *                                                                              *
 ********************************************************************************
 */

#define _OS_WAIT_EVENT_TIMEOUT(event, timeout)      \
{                                                   \
    _OS_InitDelay(timeout);                         \
    _OS_ClearReadySetCanContinue();                 \
    _OS_CHECK_EVENT((event) != 0);      /*91103*/   \
}







/************************************************************************/
/*  Wait for condition                                                  */
/************************************************************************/

#define OS_Cond_Wait(condition)     _OS_WAIT_EVENT(condition)
#define OS_Wait(condition)           OS_Cond_Wait(condition)                        // *91106.2*


/************************************************************************/
/*  Wait for condition. Exit if timeout expired.                        */
/************************************************************************/

#define OS_Cond_Wait_TO(condition, timeout)    _OS_WAIT_EVENT_TIMEOUT(condition, timeout)
#define OS_Wait_TO(condition, timeout)          OS_Cond_Wait_TO(condition, timeout) // *91106.2*



