#ifdef __PCM__
#include <16F628a.H>
#include "ccs_device.h"
#endif

#ifdef __PCH__
#include <18F252.H>
#include "ccs_device.h"
#endif

#include <osa.h>

#ifdef __OSAPICC16__
#include <pic.h>
#endif

#ifdef __OSAPICC18__
#include <pic18.h>
#endif

#ifdef __OSAMCC18__
#include <p18f242.h>
#endif


//******************************************************************************
//  Variables
//******************************************************************************

OST_QUEUE           m_queue;
OST_SQUEUE          m_squeue;
OST_SMSG            m_smsg;
OST_MSG_CB          m_msg_cb;
OST_MSG             m_msg_2;
OST_FLAG            m_flag;
OST_DTIMER          m_dt1, m_dt2;
OST_CSEM            m_csem;
OST_TASK_POINTER    m_task;


OST_MSG             msgbuf[3];


__osa_uint8         m_cCounterRec[10];
__osa_uint32        m_dwTime;

#ifdef __OSACCS__
const char DATA0[] = "aaa";
const char DATA1[] = "bbb";
#else
const char DATA[][4] = {"aaa", "bbb", "ccc", "ddd"};
#endif

const char DATA_TEXT[] = "Data";

//******************************************************************************
//  Prototypes
//******************************************************************************

#ifdef __OSAMIKROC__
#pragma funcall main Task_Send
#pragma funcall main Task_Task0
#pragma funcall main Task_Task1
#pragma funcall main Task_Rec0
#pragma funcall main Task_Rec1
#pragma funcall main Task_Rec2
#pragma funcall main Task_Rec_Csem
#pragma funcall main Task_Rec_Queue
#endif

void Task_Task0 (void);
void Task_Task1 (void);
void Task_Rec0 (void);
void Task_Rec1 (void);
void Task_Rec2 (void);
void Task_Rec_Csem (void);
void Task_Rec_Queue (void);

//******************************************************************************
//   
//   Interrupt services 
//  
//******************************************************************************

#ifndef __OSAPICC12__

#if           defined(__OSAPICC16__)
    void interrupt isr (void)
#elif         defined(__OSACCS__)
        //#INT_TIMER0
    void timer0_isr (void)
#elif        defined(__OSAMIKROC__)
    void interrupt (void)
#else
    void interrupt (void)
#endif
{
    OS_Squeue_Send_I(m_squeue, 0);
    OS_Queue_Send_I(m_queue, 0);
    OS_Msg_Send_I(m_msg_cb, 0);
    OS_Smsg_Send_I(m_smsg, 0);
    OS_Csem_Signal_I(m_csem);
}

#endif

/********************************************************************************
 *
 *  TASK: Task_Send
 *
 *------------------------------------------------------------------------------
 *
 *  description:    Sending all events
 *
 ********************************************************************************/

void Task_Send (void)
{
    static char s_cCounter;

    s_cCounter = 0;

    for (;;)
    {
        s_cCounter ++;

        OS_EnterCriticalSection();
        OS_Bsem_Set(0);
        OS_Bsem_Set(20);
        OS_LeaveCriticalSection();


        OSM_BEGIN {
            __OS_MSG_DI();
            while (OS_Msg_Check(m_msg_cb))
            {
                __OS_MSG_RI();
                _OS_WAIT_EVENT(!OS_Msg_Check(m_msg_cb));
                __OS_MSG_DI();
            };
            (m_msg_cb).status = 1;
            (m_msg_cb).msg = OST_CONVERT_TYPE_MSG (DATA_TEXT);
//            (m_msg_cb).msg =  (DATA_TEXT);
            __OS_MSG_RI();
        } OSM_END;

//        OS_Msg_Send(m_msg_cb, DATA_TEXT);

        OS_Smsg_Send(m_smsg, s_cCounter);

        #ifdef __OSACCS__
        OS_Queue_Send(m_queue, s_cCounter & 1? DATA0:DATA1);
        #else
        OS_Queue_Send(m_queue, DATA[s_cCounter & 3]);
        #endif

        OS_Csem_Signal(m_csem);

        OS_Yield();
    }
}


//******************************************************************************
//  
//  Timers and delays
// 
//******************************************************************************


//******************************************************************************
//  
//  Tasks
// 
//******************************************************************************

void Task_Rec0 (void)
{
    m_cCounterRec[0] = 0;
    for (;;)
    {
        OS_Bsem_Wait(0);
        m_cCounterRec[0]++;
    }    
}    

void Task_Rec1 (void)
{
    m_cCounterRec[1] = 0;
    for (;;)
    {
        OS_Bsem_Wait(20);
        m_cCounterRec[1]++;
    }    
}    

void Task_Rec2 (void)
{
    OST_SMSG smsg;
    m_cCounterRec[2] = 0;
    for (;;)
    {
        OS_Smsg_Wait(m_smsg, smsg);
        OS_Delay(smsg);
        m_cCounterRec[2]++;
    }    
}    

//******************************************************************************
void Task_Rec_Csem (void)
{
    for (;;)
    {
        OS_Csem_Wait(m_csem);
        OS_Delay(20);
        if (m_task == 0) OS_Task_Create(6, Task_Rec_Queue);
        m_cCounterRec[3]++;
    }
}


//******************************************************************************
void Task_Task1 (void)
{
    for (;;)
    {
        OS_Task_Replace(7, Task_Task0);
    }    
}
    static volatile OST_MSG msg;

void Task_Task0 (void)
{
    OST_CODE_POINTER cp;
    
    OS_Msg_Wait(m_msg_cb, msg);
    msg = m_msg_cb.msg;
    if (msg == DATA_TEXT)
    {
        cp = (OST_CODE_POINTER)&Task_Task1;
        OS_Task_Create_P(7, cp);
    }
    OS_Task_Stop();
}    

//******************************************************************************

void Task_Rec_Queue (void)
{
    OST_MSG msg;
    char p[4];
    char a;
    
    m_task = OS_GetCurTask();
    for (;;)
    {
        OS_Queue_Wait(m_queue, msg);
        p[0] = *msg;
        p[1] = *msg + 1;
        p[2] = *msg + 2;
        p[3] = *msg + 3;
    }    
}        


//******************************************************************************
//  
//  Waiting events
//  
//******************************************************************************




void main (void)
{
    OS_Init();

    #ifdef __OSACCS__
        OS_Task_Reserve(Task_Send);
        OS_Task_Reserve(Task_Rec0);
        OS_Task_Reserve(Task_Rec1);
        OS_Task_Reserve(Task_Rec2);
        OS_Task_Reserve(Task_Rec_Csem);
        OS_Task_Reserve(Task_Task0);
    #endif
    OS_Task_Reserve(Task_Task1);
    OS_Task_Reserve(Task_Rec_Queue);

    OS_Task_Create(7, Task_Send);
    OS_Task_Create(0, Task_Rec0);
    OS_Task_Create(1, Task_Rec1);
    OS_Task_Create(2, Task_Rec2);
    OS_Task_Create(0, Task_Rec_Csem);
    OS_Task_Create(7, Task_Task0);


    OS_Queue_Create(m_queue, msgbuf, 3);
    OS_Csem_Create(m_csem);
    OS_Msg_Create(m_msg_cb);

    m_task = 0;

    OS_EI();

    for (;;)
    {
        OS_Sched();
        OS_Timer();
        m_dwTime++;
        OS_CLRWDT();
    }
        
}
